@@ -8,7 +8,7 @@
 
 // clang-format off
 // Splitted into 2 lists because of the limitation of our macros
-#define ATOM_LIST \
+#define ATOM_LIST1 \
 	_NET_WM_WINDOW_OPACITY, \
 	_NET_FRAME_EXTENTS, \
 	WM_STATE, \
@@ -16,12 +16,16 @@
 	_NET_WM_PID, \
 	WM_NAME, \
 	WM_CLASS, \
+	WM_ICON_NAME, \
 	WM_TRANSIENT_FOR, \
 	WM_WINDOW_ROLE, \
 	WM_CLIENT_LEADER, \
+	WM_CLIENT_MACHINE, \
 	_NET_ACTIVE_WINDOW, \
 	_COMPTON_SHADOW, \
-	_NET_WM_WINDOW_TYPE, \
+	_NET_WM_WINDOW_TYPE
+
+#define ATOM_LIST2 \
 	_NET_WM_WINDOW_TYPE_DESKTOP, \
 	_NET_WM_WINDOW_TYPE_DOCK, \
 	_NET_WM_WINDOW_TYPE_TOOLBAR, \
@@ -38,14 +42,17 @@
 	_NET_WM_WINDOW_TYPE_DND, \
 	_NET_WM_STATE, \
 	_NET_WM_STATE_FULLSCREEN, \
-	_NET_WM_BYPASS_COMPOSITOR
+	_NET_WM_BYPASS_COMPOSITOR, \
+	UTF8_STRING, \
+	C_STRING
 // clang-format on
 
 #define ATOM_DEF(x) xcb_atom_t a##x
 
 struct atom {
 	struct cache *c;
-	LIST_APPLY(ATOM_DEF, SEP_COLON, ATOM_LIST);
+	LIST_APPLY(ATOM_DEF, SEP_COLON, ATOM_LIST1);
+	LIST_APPLY(ATOM_DEF, SEP_COLON, ATOM_LIST2);
 };
 
 struct atom *init_atoms(xcb_connection_t *);
