@@ -302,16 +302,21 @@ static backend_t *glx_init(session_t *ps) {
 			continue;
 		}
 
-		gd->ctx = glXCreateContextAttribsARB(ps->dpy, cfg[i], 0, true,
-		                                     (int[]){
-		                                         GLX_CONTEXT_MAJOR_VERSION_ARB,
-		                                         3,
-		                                         GLX_CONTEXT_MINOR_VERSION_ARB,
-		                                         3,
-		                                         GLX_CONTEXT_PROFILE_MASK_ARB,
-		                                         GLX_CONTEXT_CORE_PROFILE_BIT_ARB,
-		                                         0,
-		                                     });
+		int *attributes = (int[]){GLX_CONTEXT_MAJOR_VERSION_ARB,
+		                          3,
+		                          GLX_CONTEXT_MINOR_VERSION_ARB,
+		                          3,
+		                          GLX_CONTEXT_PROFILE_MASK_ARB,
+		                          GLX_CONTEXT_CORE_PROFILE_BIT_ARB,
+		                          0,
+		                          0,
+		                          0};
+		if (glxext.has_GLX_ARB_create_context_robustness) {
+			attributes[6] = GLX_CONTEXT_RESET_NOTIFICATION_STRATEGY_ARB;
+			attributes[7] = GLX_LOSE_CONTEXT_ON_RESET_ARB;
+		}
+
+		gd->ctx = glXCreateContextAttribsARB(ps->dpy, cfg[i], 0, true, attributes);
 		free(cfg);
 
 		if (!gd->ctx) {
@@ -388,11 +393,10 @@ glx_bind_pixmap(backend_t *base, xcb_pixmap_t pixmap, struct xvisual_info fmt, b
 	}
 
 	log_trace("Binding pixmap %#010x", pixmap);
-	auto wd = ccalloc(1, struct backend_image);
-	wd->max_brightness = 1;
+	auto wd = default_new_backend_image(r->width, r->height);
 	auto inner = ccalloc(1, struct gl_texture);
-	inner->width = wd->ewidth = r->width;
-	inner->height = wd->eheight = r->height;
+	inner->width = r->width;
+	inner->height = r->height;
 	wd->inner = (struct backend_image_inner_base *)inner;
 	free(r);
 
@@ -440,9 +444,6 @@ glx_bind_pixmap(backend_t *base, xcb_pixmap_t pixmap, struct xvisual_info fmt, b
 	inner->user_data = glxpixmap;
 	inner->texture = gl_new_texture(GL_TEXTURE_2D);
 	inner->has_alpha = fmt.alpha_size != 0;
-	wd->opacity = 1;
-	wd->color_inverted = false;
-	wd->dim = 0;
 	wd->inner->refcount = 1;
 	glBindTexture(GL_TEXTURE_2D, inner->texture);
 	glXBindTexImageEXT(gd->display, glxpixmap->glpixmap, GLX_FRONT_LEFT_EXT, NULL);
@@ -494,7 +495,7 @@ static void glx_diagnostics(backend_t *base) {
 
 	auto gl_renderer = (const char *)glGetString(GL_RENDERER);
 	printf("* GL renderer: %s\n", gl_renderer);
-	if (strcmp(glx_vendor, "Mesa Project and SGI")) {
+	if (strcmp(glx_vendor, "Mesa Project and SGI") == 0) {
 		for (size_t i = 0; i < ARR_SIZE(software_renderer_names); i++) {
 			if (strstr(gl_renderer, software_renderer_names[i]) != NULL) {
 				warn_software_rendering = true;
@@ -529,18 +530,26 @@ struct backend_operations glx_ops = {
     .release_image = gl_release_image,
     .compose = gl_compose,
     .image_op = gl_image_op,
-    .set_image_property = default_set_image_property,
+    .set_image_property = gl_set_image_property,
     .clone_image = default_clone_image,
     .blur = gl_blur,
     .is_image_transparent = default_is_image_transparent,
     .present = glx_present,
     .buffer_age = glx_buffer_age,
-    .render_shadow = default_backend_render_shadow,
+    .create_shadow_context = gl_create_shadow_context,
+    .destroy_shadow_context = gl_destroy_shadow_context,
+    .render_shadow = backend_render_shadow_from_mask,
+    .shadow_from_mask = gl_shadow_from_mask,
+    .make_mask = gl_make_mask,
     .fill = gl_fill,
     .create_blur_context = gl_create_blur_context,
     .destroy_blur_context = gl_destroy_blur_context,
     .get_blur_size = gl_get_blur_size,
     .diagnostics = glx_diagnostics,
+    .device_status = gl_device_status,
+    .create_shader = gl_create_window_shader,
+    .destroy_shader = gl_destroy_window_shader,
+    .get_shader_attributes = gl_get_shader_attributes,
     .max_buffer_age = 5,        // Why?
 };
 
@@ -608,6 +617,7 @@ void glxext_init(Display *dpy, int screen) {
 	check_ext(GLX_EXT_texture_from_pixmap);
 	check_ext(GLX_ARB_create_context);
 	check_ext(GLX_EXT_buffer_age);
+	check_ext(GLX_ARB_create_context_robustness);
 #ifdef GLX_MESA_query_renderer
 	check_ext(GLX_MESA_query_renderer);
 #endif
