@@ -3,13 +3,21 @@
 // Copyright (c) 2013 Richard Grenville <pyxlcy@gmail.com>
 
 #include <ctype.h>
+#include <errno.h>
+#include <fcntl.h>
 #include <limits.h>
 #include <math.h>
 #include <stdbool.h>
+#include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/stat.h>
+#include <sys/types.h>
+#include <unistd.h>
 #include <xcb/render.h>        // for xcb_render_fixed_t, XXX
 
+#include <test.h>
+
 #include "c2.h"
 #include "common.h"
 #include "compiler.h"
@@ -23,6 +31,98 @@
 
 #include "config.h"
 
+const char *xdg_config_home(void) {
+	char *xdgh = getenv("XDG_CONFIG_HOME");
+	char *home = getenv("HOME");
+	const char *default_dir = "/.config";
+
+	if (!xdgh) {
+		if (!home) {
+			return NULL;
+		}
+
+		xdgh = cvalloc(strlen(home) + strlen(default_dir) + 1);
+
+		strcpy(xdgh, home);
+		strcat(xdgh, default_dir);
+	} else {
+		xdgh = strdup(xdgh);
+	}
+
+	return xdgh;
+}
+
+char **xdg_config_dirs(void) {
+	char *xdgd = getenv("XDG_CONFIG_DIRS");
+	size_t count = 0;
+
+	if (!xdgd) {
+		xdgd = "/etc/xdg";
+	}
+
+	for (int i = 0; xdgd[i]; i++) {
+		if (xdgd[i] == ':') {
+			count++;
+		}
+	}
+
+	// Store the string and the result pointers together so they can be
+	// freed together
+	char **dir_list = cvalloc(sizeof(char *) * (count + 2) + strlen(xdgd) + 1);
+	auto dirs = strcpy((char *)dir_list + sizeof(char *) * (count + 2), xdgd);
+	auto path = dirs;
+
+	for (size_t i = 0; i < count; i++) {
+		dir_list[i] = path;
+		path = strchr(path, ':');
+		*path = '\0';
+		path++;
+	}
+	dir_list[count] = path;
+
+	size_t fill = 0;
+	for (size_t i = 0; i <= count; i++) {
+		if (dir_list[i][0] == '/') {
+			dir_list[fill] = dir_list[i];
+			fill++;
+		}
+	}
+
+	dir_list[fill] = NULL;
+
+	return dir_list;
+}
+
+TEST_CASE(xdg_config_dirs) {
+	auto old_var = getenv("XDG_CONFIG_DIRS");
+	if (old_var) {
+		old_var = strdup(old_var);
+	}
+	unsetenv("XDG_CONFIG_DIRS");
+
+	auto result = xdg_config_dirs();
+	TEST_STREQUAL(result[0], "/etc/xdg");
+	TEST_EQUAL(result[1], NULL);
+	free(result);
+
+	setenv("XDG_CONFIG_DIRS", ".:.:/etc/xdg:.:/:", 1);
+	result = xdg_config_dirs();
+	TEST_STREQUAL(result[0], "/etc/xdg");
+	TEST_STREQUAL(result[1], "/");
+	TEST_EQUAL(result[2], NULL);
+	free(result);
+
+	setenv("XDG_CONFIG_DIRS", ":", 1);
+	result = xdg_config_dirs();
+	TEST_EQUAL(result[0], NULL);
+	free(result);
+
+	if (old_var) {
+		setenv("XDG_CONFIG_DIRS", old_var, 1);
+		free(old_var);
+	}
+}
+
 /**
  * Parse a long number.
  */
@@ -438,6 +538,114 @@ bool parse_rule_opacity(c2_lptr_t **res, const char *src) {
 	return c2_parse(res, endptr, (void *)val);
 }
 
+/// Search for auxiliary file under a base directory
+static char *locate_auxiliary_file_at(const char *base, const char *scope, const char *file) {
+	scoped_charp path = mstrjoin(base, scope);
+	mstrextend(&path, "/");
+	mstrextend(&path, file);
+	if (access(path, O_RDONLY) == 0) {
+		// Canonicalize path to avoid duplicates
+		char *abspath = realpath(path, NULL);
+		return abspath;
+	}
+	return NULL;
+}
+
+/**
+ * Get a path of an auxiliary file to read, could be a shader file, or any supplimenrary
+ * file.
+ *
+ * Follows the XDG specification to search for the shader file in configuration locations.
+ *
+ * The search order is:
+ *   1) If an absolute path is given, use it directly.
+ *   2) Search for the file directly under `include_dir`.
+ *   3) Search for the file in the XDG configuration directories, under path
+ *      /picom/<scope>/
+ */
+char *locate_auxiliary_file(const char *scope, const char *path, const char *include_dir) {
+	if (!path || strlen(path) == 0) {
+		return NULL;
+	}
+
+	// Filename is absolute path, so try to load from there
+	if (path[0] == '/') {
+		if (access(path, O_RDONLY) == 0) {
+			return realpath(path, NULL);
+		}
+	}
+
+	// First try to load file from the include directory (i.e. relative to the
+	// config file)
+	if (include_dir && strlen(include_dir)) {
+		char *ret = locate_auxiliary_file_at(include_dir, "", path);
+		if (ret) {
+			return ret;
+		}
+	}
+
+	// Fall back to searching in user config directory
+	scoped_charp picom_scope = mstrjoin("/picom/", scope);
+	scoped_charp config_home = (char *)xdg_config_home();
+	char *ret = locate_auxiliary_file_at(config_home, picom_scope, path);
+	if (ret) {
+		return ret;
+	}
+
+	// Fall back to searching in system config directory
+	auto config_dirs = xdg_config_dirs();
+	for (int i = 0; config_dirs[i]; i++) {
+		ret = locate_auxiliary_file_at(config_dirs[i], picom_scope, path);
+		if (ret) {
+			free(config_dirs);
+			return ret;
+		}
+	}
+	free(config_dirs);
+
+	return ret;
+}
+
+/**
+ * Parse a list of window shader rules.
+ */
+bool parse_rule_window_shader(c2_lptr_t **res, const char *src, const char *include_dir) {
+	if (!src) {
+		return false;
+	}
+
+	// Find custom shader terminator
+	const char *endptr = strchr(src, ':');
+	if (!endptr) {
+		log_error("Custom shader terminator not found: %s", src);
+		return false;
+	}
+
+	// Parse and create custom shader
+	scoped_charp untrimed_shader_source = strdup(src);
+	if (!untrimed_shader_source) {
+		return false;
+	}
+	auto source_end = strchr(untrimed_shader_source, ':');
+	*source_end = '\0';
+
+	size_t length;
+	char *tmp = (char *)trim_both(untrimed_shader_source, &length);
+	tmp[length] = '\0';
+	char *shader_source = NULL;
+
+	if (strcasecmp(tmp, "default") != 0) {
+		shader_source = locate_auxiliary_file("shaders", tmp, include_dir);
+		if (!shader_source) {
+			log_error("Custom shader file \"%s\" not found for rule: %s", tmp, src);
+			free(shader_source);
+			return false;
+		}
+	}
+
+	return c2_parse(res, ++endptr, (void *)shader_source);
+}
+
 /**
  * Add a pattern to a condition linked list.
  */
@@ -512,6 +720,7 @@ char *parse_config(options_t *opt, const char *config_file, bool *shadow_enable,
 	// clang-format off
 	*opt = (struct options){
 	    .backend = BKEND_XRENDER,
+	    .legacy_backends = false,
 	    .glx_no_stencil = false,
 	    .mark_wmwin_focused = false,
 	    .mark_ovredir_focused = false,
@@ -565,6 +774,8 @@ char *parse_config(options_t *opt, const char *config_file, bool *shadow_enable,
 	    .blur_background_blacklist = NULL,
 	    .blur_kerns = NULL,
 	    .blur_kernel_count = 0,
+	    .window_shader_fg = NULL,
+	    .window_shader_fg_rules = NULL,
 	    .inactive_dim = 0.0,
 	    .inactive_dim_fixed = false,
 	    .invert_color_list = NULL,
