@@ -95,6 +95,16 @@ bool glx_init(session_t *ps, bool need_render) {
 			ppass->unifm_offset_x = -1;
 			ppass->unifm_offset_y = -1;
 		}
+
+		ps->psglx->round_passes = ccalloc(1, glx_round_pass_t);
+		glx_round_pass_t *ppass = ps->psglx->round_passes;
+		ppass->unifm_radius = -1;
+		ppass->unifm_texcoord = -1;
+		ppass->unifm_texsize = -1;
+		ppass->unifm_borderw = -1;
+		ppass->unifm_borderc = -1;
+		ppass->unifm_resolution = -1;
+		ppass->unifm_tex_scr = -1;
 	}
 
 	glx_session_t *psglx = ps->psglx;
@@ -232,13 +242,24 @@ void glx_destroy(session_t *ps) {
 	// Free GLSL shaders/programs
 	for (int i = 0; i < ps->o.blur_kernel_count; ++i) {
 		glx_blur_pass_t *ppass = &ps->psglx->blur_passes[i];
-		if (ppass->frag_shader)
+		if (ppass->frag_shader) {
 			glDeleteShader(ppass->frag_shader);
-		if (ppass->prog)
+		}
+		if (ppass->prog) {
 			glDeleteProgram(ppass->prog);
+		}
 	}
 	free(ps->psglx->blur_passes);
 
+	glx_round_pass_t *ppass = ps->psglx->round_passes;
+	if (ppass->frag_shader) {
+		glDeleteShader(ppass->frag_shader);
+	}
+	if (ppass->prog) {
+		glDeleteProgram(ppass->prog);
+	}
+	free(ps->psglx->round_passes);
+
 	glx_free_prog_main(&ps->glx_prog_win);
 
 	gl_check_err();
@@ -415,6 +436,139 @@ bool glx_init_blur(session_t *ps) {
 	return true;
 }
 
+/**
+ * Initialize GLX rounded corners filter.
+ */
+bool glx_init_rounded_corners(session_t *ps) {
+	char *lc_numeric_old = strdup(setlocale(LC_NUMERIC, NULL));
+	// Enforce LC_NUMERIC locale "C" here to make sure decimal point is sane
+	// Thanks to hiciu for reporting.
+	setlocale(LC_NUMERIC, "C");
+
+	static const char *FRAG_SHADER =
+	    "#version 110\n"
+	    "%s"        // extensions
+	    "uniform float u_radius;\n"
+	    "uniform float u_borderw;\n"
+	    "uniform vec4 u_borderc;\n"
+	    "uniform vec2 u_texcoord;\n"
+	    "uniform vec2 u_texsize;\n"
+	    "uniform vec2 u_resolution;\n"
+	    "uniform %s tex_scr;\n"        // sampler2D | sampler2DRect
+	    "\n"
+	    "// https://www.shadertoy.com/view/ltS3zW\n"
+	    "float RectSDF(vec2 p, vec2 b, float r) {\n"
+	    "  vec2 d = abs(p) - b + vec2(r);\n"
+	    "  return min(max(d.x, d.y), 0.0) + length(max(d, 0.0)) - r;\n"
+	    "}\n\n"
+	    "void main()\n"
+	    "{\n"
+	    "  vec2 coord = vec2(u_texcoord.x, "
+	    "u_resolution.y-u_texsize.y-u_texcoord.y);\n"
+	    "  vec4 u_v4WndBgColor = %s(tex_scr, vec2(gl_TexCoord[0].st));\n"
+	    "  float u_fRadiusPx = u_radius;\n"
+	    "  float u_fHalfBorderThickness = u_borderw / 2.0;\n"
+	    "  vec4 u_v4BorderColor = u_borderc;\n"
+	    "  vec4 u_v4FillColor = vec4(0.0, 0.0, 0.0, 0.0);\n"
+	    "  vec4 v4FromColor = u_v4BorderColor;	//Always the border "
+	    "color. If no border, this still should be set\n"
+	    "  vec4 v4ToColor = u_v4WndBgColor;		//Outside color is the "
+	    "background texture\n"
+	    "\n"
+	    "  vec2 u_v2HalfShapeSizePx = u_texsize/2.0 - "
+	    "vec2(u_fHalfBorderThickness);\n"
+	    "  vec2 v_v2CenteredPos = (gl_FragCoord.xy - u_texsize.xy / 2.0 - "
+	    "coord);\n"
+	    "\n"
+	    "  float fDist = RectSDF(v_v2CenteredPos, u_v2HalfShapeSizePx, "
+	    "u_fRadiusPx - u_fHalfBorderThickness);\n"
+	    "  if (u_fHalfBorderThickness > 0.0) {\n"
+	    "    if (fDist < 0.0) {\n"
+	    "      v4ToColor = u_v4FillColor;\n"
+	    "    }\n"
+	    "    fDist = abs(fDist) - u_fHalfBorderThickness;\n"
+	    "  } else {\n"
+	    "    v4FromColor = u_v4FillColor;\n"
+	    "  }\n"
+	    "  float fBlendAmount = smoothstep(-1.0, 1.0, fDist);\n"
+	    "  vec4 c = mix(v4FromColor, v4ToColor, fBlendAmount);\n"
+	    "\n"
+	    "  // final color\n"
+	    "  gl_FragColor = c;\n"
+	    "\n"
+	    "}\n";
+
+	const bool use_texture_rect = !ps->psglx->has_texture_non_power_of_two;
+	const char *sampler_type = (use_texture_rect ? "sampler2DRect" : "sampler2D");
+	const char *texture_func = (use_texture_rect ? "texture2DRect" : "texture2D");
+	char *extension = NULL;
+	if (use_texture_rect) {
+		mstrextend(&extension, "#extension GL_ARB_texture_rectangle : "
+		                       "require\n");
+	}
+	if (!extension) {
+		extension = strdup("");
+	}
+
+	bool success = false;
+	// Build rounded corners shader
+	auto ppass = ps->psglx->round_passes;
+	auto len = strlen(FRAG_SHADER) + strlen(extension) + strlen(sampler_type) +
+	           strlen(texture_func) + 1;
+	char *shader_str = ccalloc(len, char);
+
+	sprintf(shader_str, FRAG_SHADER, extension, sampler_type, texture_func);
+	assert(strlen(shader_str) < len);
+
+	log_debug("Generated rounded corners shader:\n%s\n", shader_str);
+
+	ppass->frag_shader = gl_create_shader(GL_FRAGMENT_SHADER, shader_str);
+	free(shader_str);
+
+	if (!ppass->frag_shader) {
+		log_error("Failed to create rounded corners fragment shader.");
+		goto out;
+	}
+
+	// Build program
+	ppass->prog = gl_create_program(&ppass->frag_shader, 1);
+	if (!ppass->prog) {
+		log_error("Failed to create GLSL program.");
+		goto out;
+	}
+
+	// Get uniform addresses
+#define P_GET_UNIFM_LOC(name, target)                                                    \
+	{                                                                                \
+		ppass->target = glGetUniformLocation(ppass->prog, name);                 \
+		if (ppass->target < 0) {                                                 \
+			log_debug("Failed to get location of rounded corners uniform "   \
+			          "'" name "'. Might be troublesome.");                  \
+		}                                                                        \
+	}
+	P_GET_UNIFM_LOC("u_radius", unifm_radius);
+	P_GET_UNIFM_LOC("u_texcoord", unifm_texcoord);
+	P_GET_UNIFM_LOC("u_texsize", unifm_texsize);
+	P_GET_UNIFM_LOC("u_borderw", unifm_borderw);
+	P_GET_UNIFM_LOC("u_borderc", unifm_borderc);
+	P_GET_UNIFM_LOC("u_resolution", unifm_resolution);
+	P_GET_UNIFM_LOC("tex_scr", unifm_tex_scr);
+#undef P_GET_UNIFM_LOC
+
+	success = true;
+
+out:
+	free(extension);
+
+	// Restore LC_NUMERIC
+	setlocale(LC_NUMERIC, lc_numeric_old);
+	free(lc_numeric_old);
+
+	gl_check_err();
+
+	return success;
+}
+
 /**
  * Load a GLSL main program from shader strings.
  */
@@ -449,6 +603,88 @@ bool glx_load_prog_main(const char *vshader_str, const char *fshader_str,
 	return true;
 }
 
+static inline void glx_copy_region_to_tex(session_t *ps, GLenum tex_tgt, int basex,
+                                          int basey, int dx, int dy, int width, int height) {
+	if (width > 0 && height > 0) {
+		glCopyTexSubImage2D(tex_tgt, 0, dx - basex, dy - basey, dx,
+		                    ps->root_height - dy - height, width, height);
+	}
+}
+
+static inline GLuint glx_gen_texture(GLenum tex_tgt, int width, int height) {
+	GLuint tex = 0;
+	glGenTextures(1, &tex);
+	if (!tex) {
+		return 0;
+	}
+	glEnable(tex_tgt);
+	glBindTexture(tex_tgt, tex);
+	glTexParameteri(tex_tgt, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
+	glTexParameteri(tex_tgt, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
+	glTexParameteri(tex_tgt, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
+	glTexParameteri(tex_tgt, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
+	glTexImage2D(tex_tgt, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
+	glBindTexture(tex_tgt, 0);
+
+	return tex;
+}
+
+/**
+ * Bind an OpenGL texture and fill it with pixel data from back buffer
+ */
+bool glx_bind_texture(session_t *ps attr_unused, glx_texture_t **pptex, int x, int y,
+                      int width, int height) {
+	if (ps->o.backend != BKEND_GLX && ps->o.backend != BKEND_XR_GLX_HYBRID) {
+		return true;
+	}
+
+	glx_texture_t *ptex = *pptex;
+
+	// log_trace("Copying xy(%d %d) wh(%d %d) ptex(%p)", x, y, width, height, ptex);
+
+	// Release texture if parameters are inconsistent
+	if (ptex && ptex->texture && (ptex->width != width || ptex->height != height)) {
+		free_texture(ps, &ptex);
+	}
+
+	// Allocate structure
+	if (!ptex) {
+		ptex = ccalloc(1, glx_texture_t);
+		*pptex = ptex;
+
+		ptex->width = width;
+		ptex->height = height;
+		ptex->target = GL_TEXTURE_RECTANGLE;
+		if (ps->psglx->has_texture_non_power_of_two) {
+			ptex->target = GL_TEXTURE_2D;
+		}
+	}
+
+	// Create texture
+	if (!ptex->texture) {
+		ptex->texture = glx_gen_texture(ptex->target, width, height);
+	}
+	if (!ptex->texture) {
+		log_error("Failed to allocate texture.");
+		return false;
+	}
+
+	// Read destination pixels into a texture
+	glEnable(ptex->target);
+	glBindTexture(ptex->target, ptex->texture);
+	if (width > 0 && height > 0) {
+		glx_copy_region_to_tex(ps, ptex->target, x, y, x, y, width, height);
+	}
+
+	// Cleanup
+	glBindTexture(ptex->target, 0);
+	glDisable(ptex->target);
+
+	gl_check_err();
+
+	return true;
+}
+
 /**
  * Bind a X pixmap to an OpenGL texture.
  */
@@ -662,30 +898,6 @@ void glx_set_clip(session_t *ps, const region_t *reg) {
                                                                                          \
 	pixman_region32_fini(&reg_new);
 
-static inline GLuint glx_gen_texture(GLenum tex_tgt, int width, int height) {
-	GLuint tex = 0;
-	glGenTextures(1, &tex);
-	if (!tex)
-		return 0;
-	glEnable(tex_tgt);
-	glBindTexture(tex_tgt, tex);
-	glTexParameteri(tex_tgt, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
-	glTexParameteri(tex_tgt, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
-	glTexParameteri(tex_tgt, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
-	glTexParameteri(tex_tgt, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
-	glTexImage2D(tex_tgt, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
-	glBindTexture(tex_tgt, 0);
-
-	return tex;
-}
-
-static inline void glx_copy_region_to_tex(session_t *ps, GLenum tex_tgt, int basex,
-                                          int basey, int dx, int dy, int width, int height) {
-	if (width > 0 && height > 0)
-		glCopyTexSubImage2D(tex_tgt, 0, dx - basex, dy - basey, dx,
-		                    ps->root_height - dy - height, width, height);
-}
-
 /**
  * Blur contents in a particular region.
  *
@@ -889,6 +1101,174 @@ glx_blur_dst_end:
 	return ret;
 }
 
+// TODO(bhagwan) this is a mess and needs a more consistent way of getting the border
+// pixel I tried looking for a notify event for XCB_CW_BORDER_PIXEL (in
+// xcb_create_window()) or a way to get the pixels from xcb_render_picture_t but the
+// documentation for the xcb_xrender extension is literaly non existent...
+//
+// NOTE(yshui) There is no consistent way to get the "border" color of a X window. From
+// the WM's perspective there are multiple ways to implement window borders. Using
+// glReadPixel is probably the most reliable way.
+void glx_read_border_pixel(int root_height, int root_width, int x, int y, int width,
+                           int height, float *ppixel) {
+	assert(ppixel);
+
+	// Reset the color so the shader doesn't use it
+	ppixel[0] = ppixel[1] = ppixel[2] = ppixel[3] = -1.0F;
+
+	// First try bottom left corner past the
+	// circle radius (after the rounded corner ends)
+	auto screen_x = x;
+	auto screen_y = root_height - height - y;
+
+	// X is out of bounds
+	// move to the right side
+	if (screen_x < 0) {
+		screen_x += width;
+	}
+
+	// Y is out of bounds
+	// move to to top part
+	if (screen_y < 0) {
+		screen_y += height;
+	}
+
+	// All corners are out of bounds, give up
+	if (screen_x < 0 || screen_y < 0 || screen_x >= root_width || screen_y >= root_height) {
+		return;
+	}
+
+	// Invert Y-axis so we can query border color from texture (0,0)
+	glReadPixels(screen_x, screen_y, 1, 1, GL_RGBA, GL_FLOAT, (void *)ppixel);
+
+	log_trace("xy(%d, %d), glxy(%d %d) wh(%d %d), border_col(%.2f, %.2f, %.2f, %.2f)",
+	          x, y, screen_x, screen_y, width, height, (float)ppixel[0],
+	          (float)ppixel[1], (float)ppixel[2], (float)ppixel[3]);
+
+	gl_check_err();
+}
+
+bool glx_round_corners_dst(session_t *ps, struct managed_win *w,
+                           const glx_texture_t *ptex, int dx, int dy, int width,
+                           int height, float z, float cr, const region_t *reg_tgt) {
+	assert(ps->psglx->round_passes->prog);
+	bool ret = false;
+
+	// log_warn("dxy(%d, %d) wh(%d %d) rwh(%d %d) b(%d), f(%d)",
+	//	dx, dy, width, height, ps->root_width, ps->root_height, w->g.border_width,
+	// w->focused);
+
+	int mdx = dx, mdy = dy, mwidth = width, mheight = height;
+	log_trace("%d, %d, %d, %d", mdx, mdy, mwidth, mheight);
+
+	if (w->g.border_width > 0) {
+		glx_read_border_pixel(ps->root_height, ps->root_width, dx, dy, width,
+		                      height, &w->border_col[0]);
+	}
+
+	{
+		const glx_round_pass_t *ppass = ps->psglx->round_passes;
+		assert(ppass->prog);
+
+		glEnable(GL_BLEND);
+		glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
+
+		glUseProgram(ppass->prog);
+
+		// If caller specified a texture use it as source
+		log_trace("ptex: %p wh(%d %d) %d %d", ptex, ptex->width, ptex->height,
+		          ptex->target, ptex->texture);
+
+		glActiveTexture(GL_TEXTURE0);
+		glBindTexture(ptex->target, ptex->texture);
+
+		if (ppass->unifm_tex_scr >= 0) {
+			glUniform1i(ppass->unifm_tex_scr, (GLint)0);
+		}
+		if (ppass->unifm_radius >= 0) {
+			glUniform1f(ppass->unifm_radius, cr);
+		}
+		if (ppass->unifm_texcoord >= 0) {
+			glUniform2f(ppass->unifm_texcoord, (float)dx, (float)dy);
+		}
+		if (ppass->unifm_texsize >= 0) {
+			glUniform2f(ppass->unifm_texsize, (float)mwidth, (float)mheight);
+		}
+		if (ppass->unifm_borderw >= 0) {
+			// Don't render rounded border if we don't know the border color
+			glUniform1f(ppass->unifm_borderw,
+			            w->border_col[0] != -1. ? (GLfloat)w->g.border_width : 0);
+		}
+		if (ppass->unifm_borderc >= 0) {
+			glUniform4f(ppass->unifm_borderc, w->border_col[0],
+			            w->border_col[1], w->border_col[2], w->border_col[3]);
+		}
+		if (ppass->unifm_resolution >= 0) {
+			glUniform2f(ppass->unifm_resolution, (float)ps->root_width,
+			            (float)ps->root_height);
+		}
+
+		// Painting
+		{
+			P_PAINTREG_START(crect) {
+				// texture-local coordinates
+				auto rx = (GLfloat)(crect.x1 - dx);
+				auto ry = (GLfloat)(crect.y1 - dy);
+				auto rxe = rx + (GLfloat)(crect.x2 - crect.x1);
+				auto rye = ry + (GLfloat)(crect.y2 - crect.y1);
+				if (GL_TEXTURE_2D == ptex->target) {
+					rx = rx / (GLfloat)width;
+					ry = ry / (GLfloat)height;
+					rxe = rxe / (GLfloat)width;
+					rye = rye / (GLfloat)height;
+				}
+
+				// coordinates for the texture in the target
+				auto rdx = (GLfloat)crect.x1;
+				auto rdy = (GLfloat)(ps->root_height - crect.y1);
+				auto rdxe = (GLfloat)rdx + (GLfloat)(crect.x2 - crect.x1);
+				auto rdye = (GLfloat)rdy - (GLfloat)(crect.y2 - crect.y1);
+
+				// Invert Y if needed, this may not work as expected,
+				// though. I don't have such a FBConfig to test with.
+				ry = 1.0F - ry;
+				rye = 1.0F - rye;
+
+				// log_trace("Rect %d (i:%d): %f, %f, %f, %f -> %f, %f,
+				// %f, %f", 	ri ,ptex ? ptex->y_inverted : -1, rx, ry,
+				// rxe,
+				// rye, rdx, rdy, rdxe, rdye);
+
+				glTexCoord2f(rx, ry);
+				glVertex3f(rdx, rdy, z);
+
+				glTexCoord2f(rxe, ry);
+				glVertex3f(rdxe, rdy, z);
+
+				glTexCoord2f(rxe, rye);
+				glVertex3f(rdxe, rdye, z);
+
+				glTexCoord2f(rx, rye);
+				glVertex3f(rdx, rdye, z);
+			}
+			P_PAINTREG_END();
+		}
+
+		glUseProgram(0);
+		glDisable(GL_BLEND);
+	}
+
+	ret = true;
+
+	glBindTexture(ptex->target, 0);
+	glDisable(ptex->target);
+	glDisable(GL_BLEND);
+
+	gl_check_err();
+
+	return ret;
+}
+
 bool glx_dim_dst(session_t *ps, int dx, int dy, int width, int height, int z,
                  GLfloat factor, const region_t *reg_tgt) {
 	// It's possible to dim in glx_render(), but it would be over-complicated
@@ -1056,7 +1436,7 @@ bool glx_render(session_t *ps, const glx_texture_t *ptex, int x, int y, int dx,
 	// Painting
 	{
 		P_PAINTREG_START(crect) {
-			// XXX explain these variables
+			// texture-local coordinates
 			auto rx = (GLfloat)(crect.x1 - dx + x);
 			auto ry = (GLfloat)(crect.y1 - dy + y);
 			auto rxe = rx + (GLfloat)(crect.x2 - crect.x1);
@@ -1069,6 +1449,8 @@ bool glx_render(session_t *ps, const glx_texture_t *ptex, int x, int y, int dx,
 				rxe = rxe / (GLfloat)ptex->width;
 				rye = rye / (GLfloat)ptex->height;
 			}
+
+			// coordinates for the texture in the target
 			GLint rdx = crect.x1;
 			GLint rdy = ps->root_height - crect.y1;
 			GLint rdxe = rdx + (crect.x2 - crect.x1);
